<?php

/**
 * Created by PhpStorm.
 * User: jderosa
 * Date: 21/04/20
 * Time: 12.02
 */

namespace FiloBlu\Rma\Handlers;

use Exception;
use FiloBlu\Rma\Api\RmaExtensionAttributesRepositoryInterface;
use FiloBlu\Rma\Helper\ChangeSizeAttributeHelper;
use FiloBlu\Rma\Helper\ChangeSizeOrder;
use FiloBlu\Rma\Helper\RmaHelper;
use Magento\Rma\Api\Data\RmaInterface;
use Magento\Sales\Model\Order\Email\Sender\OrderSender;
use function is_array;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\Webapi\Rest\Request;
use Magento\Rma\Model\Item as ItemModel;
use Magento\Rma\Model\Rma\Source\Status;
use Magento\Store\Model\StoreManagerInterface;
use Psr\Log\LoggerInterface;
use Throwable;

/**
 *
 */
class CreateChangeSizeOrderHandler
{
    /**
     * @var array
     */
    protected $changSizeItems = [];
    /**
     * @var Request
     */
    protected $webApiRequest;
    /**
     * @var OrderSender
     */
    protected $orderSender;
    /**
     * @var ChangeSizeAttributeHelper
     */
    private $changeSizeAttributeHelper;
    /**
     * @var RmaExtensionAttributesRepositoryInterface
     */
    private $rmaExtensionAttributesRepository;
    /**
     * @var LoggerInterface
     */
    private $logger;

    /**
     * @var ChangeSizeOrder
     */
    private $changeSizeOrder;

    /**
     * @var RequestInterface
     */
    private $request;

    /**
     * @var StoreManagerInterface
     */
    private $storeManager;

    /**
     * @var ItemModel
     */
    private $itemModel;
    /**
     * @var RmaHelper
     */
    private $rmaHelper;


    /**
     * CreateChangeSizeOrderHandler constructor.
     * @param ChangeSizeAttributeHelper $changeSizeAttributeHelper
     * @param RmaExtensionAttributesRepositoryInterface $rmaExtensionAttributesRepository
     * @param LoggerInterface $logger
     * @param ChangeSizeOrder $changeSizeOrder
     * @param RequestInterface $request
     * @param StoreManagerInterface $storeManager
     * @param ItemModel $itemModel
     * @param RmaHelper $rmaHelper
     * @param Request $webApiRequest
     * @param OrderSender $orderSender
     */
    public function __construct(
        ChangeSizeAttributeHelper $changeSizeAttributeHelper,
        RmaExtensionAttributesRepositoryInterface $rmaExtensionAttributesRepository,
        LoggerInterface $logger,
        ChangeSizeOrder $changeSizeOrder,
        RequestInterface $request,
        StoreManagerInterface $storeManager,
        ItemModel $itemModel,
        RmaHelper $rmaHelper,
        Request $webApiRequest,
        OrderSender $orderSender
    ) {
        $this->changeSizeAttributeHelper = $changeSizeAttributeHelper;
        $this->rmaExtensionAttributesRepository = $rmaExtensionAttributesRepository;
        $this->logger = $logger;
        $this->changeSizeOrder = $changeSizeOrder;
        $this->request = $request;
        $this->storeManager = $storeManager;
        $this->itemModel = $itemModel;
        $this->rmaHelper = $rmaHelper;
        $this->webApiRequest = $webApiRequest;
        $this->orderSender = $orderSender;
    }

    /**
     * TODO: Refactor
     * @param RmaInterface $subject
     * @param array $result
     * @return array
     * @throws \Exception
     */
    public function checkIfCreateChangeSizeOrder(RmaInterface $subject, RmaInterface $result)
    {
        $websiteId = $this->storeManager->getStore($subject->getStoreId())->getWebsiteId();

        if (!$this->changeSizeAttributeHelper->isChangeSizeEnabled($websiteId)) {
            return $result;
        }

        if (!$result || !isset($result['items']) || !is_array($result['items'])) {
            return $result;
        }

        try {
            $firstItem = true;
            $calculateShipping = true;
            /** @var \Magento\Rma\Api\Data\RmaInterface[] $rmaItems */
            $rmaItems = $result->getItems();
            $parentProduct = null;
            $orderCreationStrategy = $this->rmaHelper->getChangeSizeOrderCreationStrategy();

            foreach ($rmaItems as $item) {
                if ($item->getStatus() !== Status::STATE_PENDING) {
                    continue;
                }

                if (
                    (int)$item->getResolution() !== $this->changeSizeAttributeHelper->getChangeSizeAttributeOptionId(
                    )
                ) {
                    continue;
                }

                if ($this->rmaExtensionAttributesRepository->getChangeSizeOrderId($item->getId())) {
                    continue;
                }

                $changeSizeItem = [];

                // check if customer
                if (
                    $item->getExtensionAttributes() !== null && $item->getExtensionAttributes()->getChangeSize(
                    ) != null
                ) {
                    $item->setChangeSize($item->getExtensionAttributes()->getChangeSize());
                }

                if (
                    $item->getChangeSize() != '' || !empty(
                        $item->getData(
                            ChangeSizeAttributeHelper::CHANGE_SIZE_RMA_ATTRIBUTE_CODE
                        )
                    )
                ) {
                    $changeProductId = $item->getChangeSize();
                    /** @var ItemModel $item */
                    $parentProduct = $item->getProductSku() ?? $this->itemModel->setEntityId(
                        $item->getEntityId()
                    )->getProductSku();
                    $changeSizeItem = [
                        'product_id'             => $changeProductId,
                        'qty'                    => $item->getQtyRequested(), // $item['qty_requested'],
                        'parent_product_sku'     => $parentProduct,
                        'rma_entity_id'          => $item->getEntityId(),
                        'original_order_item_id' => $item->getOrderItemId()
                    ];
                }

                $customEmail = $subject->getCustomerCustomEmail();
                if (isset($customEmail) && $customEmail !== '') {
                    $subject->getOrder()->setCustomerEmail($customEmail);
                }

                if ($orderCreationStrategy === 'each') {
                    // Create an order for each item requested
                    $order = $this->createChangeSizeOrder(
                        $subject->getOrder(),
                        [$changeSizeItem],
                        $subject->getEntityId(),
                        $calculateShipping,
                        $parentProduct
                    );

                    if ($firstItem) {
                        $firstItem = false;
                        $calculateShipping = false;
                    }
                    $this->rmaExtensionAttributesRepository->setChangeSizeOrderId($item->getEntityId(), $order->getId());
                    $this->orderSender->send($order);

                    continue;
                }

                $this->changSizeItems[] = $changeSizeItem;
            }

            if ($orderCreationStrategy === 'total' && $this->changSizeItems) {
                //create on order with all items requested
                $this->createChangeSizeOrder(
                    $subject->getOrder(),
                    $this->changSizeItems,
                    $subject->getEntityId(),
                    $calculateShipping,
                    $parentProduct
                );
            }
        } catch (Exception $exception) {
            $this->logger->critical($exception->getMessage(), ['exception' => $exception]);
            throw new Exception($exception->getMessage());
        } catch (Throwable $throwable) {
            $this->logger->critical($throwable->getMessage(), ['exception' => $throwable]);
            throw new Exception($throwable->getMessage());
        }
        return $result;
    }

    /**
     * @param \Magento\Sales\Model\Order| \Magento\Sales\Api\Data\OrderInterface $order
     * @param array $items
     * @param $rmaId
     * @param null $parentProduct
     * @param bool $calculateShipping
     * @return bool|\Magento\Sales\Api\Data\OrderInterface
     * @throws \Exception
     */
    protected function createChangeSizeOrder($order, array $items, $rmaId, bool $calculateShipping, $parentProduct = null)
    {
        $orderData = [
            'order'              => $order,
            'rma_id'             => $rmaId,
            'parent_product_sku' => $parentProduct,
            'items'              => $items,
            'request'            => $this->request->getParams(),
            'calculate_shipping_cost' => $calculateShipping
        ];

        $request = $this->request->getParams();

        if (!$request) {
            $request = $this->webApiRequest->getParams();
        }

        if (!$request) {
            $request = $this->webApiRequest->getBodyParams();
        }

        if (isset($request['rmaDataObject']['extension_attributes'])) {
            $requestData = $request['rmaDataObject']['extension_attributes'];
            $orderData['request'] = array_merge($orderData['request'], $requestData);
            $orderData['request'] = array_merge($orderData['request'], $requestData['pickup_address']);
            unset($orderData['pickup_address']);
        }

        return $this->changeSizeOrder->createMageOrder($orderData);
    }
}
